@@ -14,13 +14,13 @@
 
 import collections
 import functools
+import logging
 import os
 import pwd
 import re
 import string
 
-from cloudinit import log as logging
-from cloudinit import net, sources, subp, util
+from cloudinit import atomic_helper, net, sources, subp, util
 
 LOG = logging.getLogger(__name__)
 
@@ -435,7 +435,7 @@ def read_context_disk_dir(source_dir, distro, asuser=None):
                 ) from e
         try:
             path = os.path.join(source_dir, "context.sh")
-            content = util.load_file(path)
+            content = util.load_text_file(path)
             context = parse_shell_config(content, asuser=asuser)
         except subp.ProcessExecutionError as e:
             raise BrokenContextDiskDir(
@@ -486,7 +486,7 @@ def read_context_disk_dir(source_dir, distro, asuser=None):
         )
         if encoding == "base64":
             try:
-                results["userdata"] = util.b64d(results["userdata"])
+                results["userdata"] = atomic_helper.b64d(results["userdata"])
             except TypeError:
                 LOG.warning("Failed base64 decoding of userdata")
 
@@ -520,6 +520,3 @@ datasources = [
 # Return a list of data sources that match this set of dependencies
 def get_datasource_list(depends):
     return sources.list_from_depends(depends, datasources)
-
-
-# vi: ts=4 expandtab
