@@ -9,8 +9,6 @@
 # This file is part of cloud-init. See LICENSE file for license information.
 
 import copy
-import ftplib
-import io
 import json
 import logging
 import os
@@ -18,32 +16,21 @@ import threading
 import time
 from concurrent.futures import ThreadPoolExecutor, TimeoutError, as_completed
 from email.utils import parsedate
+from errno import ENOENT
 from functools import partial
 from http.client import NOT_FOUND
 from itertools import count
-from ssl import create_default_context
-from typing import (
-    Any,
-    Callable,
-    Iterator,
-    List,
-    Mapping,
-    NamedTuple,
-    Optional,
-    Tuple,
-    Union,
-)
-from urllib.parse import quote, urlparse, urlsplit, urlunparse
+from typing import Any, Callable, Iterator, List, Optional, Tuple, Union
+from urllib.parse import quote, urlparse, urlunparse
 
 import requests
 from requests import exceptions
 
-from cloudinit import performance, util, version
+from cloudinit import version
 
 LOG = logging.getLogger(__name__)
 
 REDACTED = "REDACTED"
-ExceptionCallback = Optional[Callable[["UrlError"], bool]]
 
 
 def _cleanurl(url):
@@ -72,178 +59,7 @@ def combine_url(base, *add_ons):
     return url
 
 
-def ftp_get_return_code_from_exception(exc) -> int:
-    """helper for read_ftps to map return codes to a number"""
-    # ftplib doesn't expose error codes, so use this lookup table
-    ftp_error_codes = {
-        ftplib.error_reply: 300,  # unexpected [123]xx reply
-        ftplib.error_temp: 400,  # 4xx errors
-        ftplib.error_perm: 500,  # 5xx errors
-        ftplib.error_proto: 600,  # response does not begin with [1-5]
-        EOFError: 700,  # made up
-        # OSError is also possible. Use OSError.errno for that.
-    }
-    code = ftp_error_codes.get(type(exc))  # pyright: ignore
-    if not code:
-        if isinstance(exc, OSError):
-            code = exc.errno
-        else:
-            LOG.warning(
-                "Unexpected exception type while connecting to ftp server."
-            )
-            code = -99
-    return code
-
-
-def read_ftps(url: str, timeout: float = 5.0, **kwargs: dict) -> "FtpResponse":
-    """connect to URL using ftp over TLS and read a file
-
-    when using strict mode (ftps://), raise exception in event of failure
-    when not using strict mode (ftp://), fall back to using unencrypted ftp
-
-    url: string containing the desination to read a file from. The url is
-        parsed with urllib.urlsplit to identify username, password, host,
-        path, and port in the following format:
-            ftps://[username:password@]host[:port]/[path]
-        host is the only required component
-    timeout: maximum time for the connection to take
-    kwargs: unused, for compatibility with read_url
-    returns: UrlResponse
-    """
-
-    url_parts = urlsplit(url)
-    if not url_parts.hostname:
-        raise UrlError(
-            cause="Invalid url provided", code=NOT_FOUND, headers=None, url=url
-        )
-    with io.BytesIO() as buffer:
-        port = url_parts.port or 21
-        user = url_parts.username or "anonymous"
-        if "ftps" == url_parts.scheme:
-            try:
-                ftp_tls = ftplib.FTP_TLS(
-                    context=create_default_context(),
-                )
-                LOG.debug(
-                    "Attempting to connect to %s via port [%s] over tls.",
-                    url,
-                    port,
-                )
-                ftp_tls.connect(
-                    host=url_parts.hostname,
-                    port=port,
-                    timeout=timeout or 5.0,  # uses float internally
-                )
-            except ftplib.all_errors as e:
-                code = ftp_get_return_code_from_exception(e)
-                raise UrlError(
-                    cause=(
-                        "Reading file from server over tls "
-                        f"failed for url {url} [{code}]"
-                    ),
-                    code=code,
-                    headers=None,
-                    url=url,
-                ) from e
-            LOG.debug("Attempting to login with user [%s]", user)
-            try:
-                ftp_tls.login(
-                    user=user,
-                    passwd=url_parts.password or "",
-                )
-                LOG.debug("Creating a secure connection")
-                ftp_tls.prot_p()
-            except ftplib.error_perm as e:
-                LOG.warning(
-                    "Attempted to connect to an insecure ftp server but used "
-                    "a scheme of ftps://, which is not allowed. Use ftp:// "
-                    "to allow connecting to insecure ftp servers."
-                )
-                raise UrlError(
-                    cause=(
-                        "Attempted to connect to an insecure ftp server but "
-                        "used a scheme of ftps://, which is not allowed. Use "
-                        "ftp:// to allow connecting to insecure ftp servers."
-                    ),
-                    code=500,
-                    headers=None,
-                    url=url,
-                ) from e
-            try:
-                LOG.debug("Reading file: %s", url_parts.path)
-                ftp_tls.retrbinary(
-                    f"RETR {url_parts.path}", callback=buffer.write
-                )
-
-                return FtpResponse(buffer.getvalue(), url)
-            except ftplib.all_errors as e:
-                code = ftp_get_return_code_from_exception(e)
-                raise UrlError(
-                    cause=(
-                        "Reading file from ftp server"
-                        f" failed for url {url} [{code}]"
-                    ),
-                    code=code,
-                    headers=None,
-                    url=url,
-                ) from e
-            finally:
-                LOG.debug("Closing connection")
-                ftp_tls.close()
-        else:
-            try:
-                ftp = ftplib.FTP()
-                LOG.debug(
-                    "Attempting to connect to %s via port %s.", url, port
-                )
-                ftp.connect(
-                    host=url_parts.hostname,
-                    port=port,
-                    timeout=timeout or 5.0,  # uses float internally
-                )
-                LOG.debug("Attempting to login with user [%s]", user)
-                ftp.login(
-                    user=user,
-                    passwd=url_parts.password or "",
-                )
-                LOG.debug("Reading file: %s", url_parts.path)
-                ftp.retrbinary(f"RETR {url_parts.path}", callback=buffer.write)
-                return FtpResponse(buffer.getvalue(), url)
-            except ftplib.all_errors as e:
-                code = ftp_get_return_code_from_exception(e)
-                raise UrlError(
-                    cause=(
-                        "Reading file from ftp server"
-                        f" failed for url {url} [{code}]"
-                    ),
-                    code=code,
-                    headers=None,
-                    url=url,
-                ) from e
-            finally:
-                LOG.debug("Closing connection")
-                ftp.close()
-
-
-def _read_file(path: str, **kwargs) -> "FileResponse":
-    """read a binary file and return a FileResponse
-
-    matches function signature with read_ftps and read_url
-    """
-    if kwargs.get("data"):
-        LOG.warning("Unable to post data to file resource %s", path)
-    try:
-        contents = util.load_binary_file(path)
-        return FileResponse(contents, path)
-    except FileNotFoundError as e:
-        raise UrlError(cause=e, code=NOT_FOUND, headers=None, url=path) from e
-    except IOError as e:
-        raise UrlError(cause=e, code=e.errno, headers=None, url=path) from e
-
-
-def read_file_or_url(
-    url, **kwargs
-) -> Union["FileResponse", "UrlResponse", "FtpResponse"]:
+def read_file_or_url(url, **kwargs) -> Union["FileResponse", "UrlResponse"]:
     """Wrapper function around readurl to allow passing a file path as url.
 
     When url is not a local file path, passthrough any kwargs to readurl.
@@ -252,19 +68,22 @@ def read_file_or_url(
     parameters. See: call-signature of readurl in this module for param docs.
     """
     url = url.lstrip()
-    try:
-        parsed = urlparse(url)
-    except ValueError as e:
-        raise UrlError(cause=e, url=url) from e
-    scheme = parsed.scheme
-    if scheme == "file" or (url and "/" == url[0]):
-        return _read_file(parsed.path, **kwargs)
-    elif scheme in ("ftp", "ftps"):
-        return read_ftps(url, **kwargs)
-    elif scheme in ("http", "https"):
-        return readurl(url, **kwargs)
+    if url.startswith("/"):
+        url = "file://%s" % url
+    if url.lower().startswith("file://"):
+        if kwargs.get("data"):
+            LOG.warning("Unable to post data to file resource %s", url)
+        file_path = url[len("file://") :]
+        try:
+            with open(file_path, "rb") as fp:
+                contents = fp.read()
+        except IOError as e:
+            code = e.errno
+            if e.errno == ENOENT:
+                code = NOT_FOUND
+            raise UrlError(cause=e, code=code, headers=None, url=url) from e
+        return FileResponse(file_path, contents=contents)
     else:
-        LOG.warning("Attempting unknown protocol %s", scheme)
         return readurl(url, **kwargs)
 
 
@@ -272,27 +91,25 @@ def read_file_or_url(
 # read_file_or_url can return this or that object and the
 # 'user' of those objects will not need to know the difference.
 class StringResponse:
-    def __init__(self, contents, url, code=200):
+    def __init__(self, contents, code=200):
         self.code = code
         self.headers = {}
         self.contents = contents
-        self.url = url
+        self.url = None
 
     def ok(self, *args, **kwargs):
-        return self.code == 200
+        if self.code != 200:
+            return False
+        return True
 
     def __str__(self):
         return self.contents.decode("utf-8")
 
 
 class FileResponse(StringResponse):
-    def __init__(self, contents: bytes, url: str, code=200):
-        super().__init__(contents, url, code=code)
-
-
-class FtpResponse(StringResponse):
-    def __init__(self, contents: bytes, url: str):
-        super().__init__(contents, url)
+    def __init__(self, path, contents, code=200):
+        StringResponse.__init__(self, contents, code=code)
+        self.url = path
 
 
 class UrlResponse:
@@ -306,10 +123,10 @@ class UrlResponse:
         return self._response.content
 
     @property
-    def url(self) -> str:
+    def url(self):
         return self._response.url
 
-    def ok(self, redirects_ok=False) -> bool:
+    def ok(self, redirects_ok=False):
         upper = 300
         if redirects_ok:
             upper = 400
@@ -323,7 +140,7 @@ class UrlResponse:
         return self._response.headers
 
     @property
-    def code(self) -> int:
+    def code(self):
         return self._response.status_code
 
     def __str__(self):
@@ -345,17 +162,13 @@ class UrlResponse:
 
 
 class UrlError(IOError):
-    def __init__(
-        self,
-        cause: Any,  # This SHOULD be an exception to wrap, but can be anything
-        code: Optional[int] = None,
-        headers: Optional[Mapping] = None,
-        url: Optional[str] = None,
-    ):
+    def __init__(self, cause, code=None, headers=None, url=None):
         IOError.__init__(self, str(cause))
         self.cause = cause
         self.code = code
-        self.headers: Mapping = {} if headers is None else headers
+        self.headers = headers
+        if self.headers is None:
+            self.headers = {}
         self.url = url
 
 
@@ -377,77 +190,8 @@ def _get_ssl_args(url, ssl_details):
     return ssl_args
 
 
-def _get_retry_after(retry_after: str) -> float:
-    """Parse a Retry-After header value into an integer.
-
-    : param retry_after: The value of the Retry-After header.
-        https://www.rfc-editor.org/rfc/rfc9110.html#section-10.2.3
-        https://www.rfc-editor.org/rfc/rfc2616#section-3.3
-    : return: The number of seconds to wait before retrying the request.
-    """
-    try:
-        to_wait = float(retry_after)
-    except ValueError:
-        # Translate a date such as "Fri, 31 Dec 1999 23:59:59 GMT"
-        # into seconds to wait
-        try:
-            time_tuple = parsedate(retry_after)
-            if not time_tuple:
-                raise ValueError("Failed to parse Retry-After header value")
-            to_wait = float(time.mktime(time_tuple) - time.time())
-        except ValueError:
-            LOG.info(
-                "Failed to parse Retry-After header value: %s. "
-                "Waiting 1 second instead.",
-                retry_after,
-            )
-            to_wait = 1
-        if to_wait < 0:
-            LOG.info(
-                "Retry-After header value is in the past. "
-                "Waiting 1 second instead."
-            )
-            to_wait = 1
-    return to_wait
-
-
-def _handle_error(
-    error: UrlError,
-    *,
-    exception_cb: ExceptionCallback = None,
-) -> Optional[float]:
-    """Handle exceptions raised during request processing.
-
-    If we have no exception callback or the callback handled the error or we
-    got a 503, return with an optional timeout so the request can be retried.
-    Otherwise, raise the error.
-
-    :param error: The exception raised during the request.
-    :param response: The response object.
-    :param exception_cb: Callable to handle the exception.
-
-    :return: Optional time to wait before retrying the request.
-    """
-    if exception_cb and exception_cb(error):
-        return None
-    if error.code and error.code == 503:
-        LOG.warning(
-            "Ec2 IMDS endpoint returned a 503 error. "
-            "HTTP endpoint is overloaded. Retrying."
-        )
-        if error.headers:
-            return _get_retry_after(error.headers.get("Retry-After", "1"))
-        LOG.info("Unable to introspect response header. Waiting 1 second.")
-        return 1
-    if not exception_cb:
-        return None
-    # If exception_cb returned False and there's no 503
-    raise error
-
-
 def readurl(
     url,
-    *,
     data=None,
     timeout=None,
     retries=0,
@@ -458,7 +202,7 @@ def readurl(
     ssl_details=None,
     check_status=True,
     allow_redirects=True,
-    exception_cb: ExceptionCallback = None,
+    exception_cb=None,
     session=None,
     infinite=False,
     log_req_resp=True,
@@ -487,8 +231,8 @@ def readurl(
         occurs. Default: True.
     :param allow_redirects: Optional boolean passed straight to Session.request
         as 'allow_redirects'. Default: True.
-    :param exception_cb: Optional callable to handle exception and returns
-        True if retries are permitted.
+    :param exception_cb: Optional callable which accepts the params
+        msg and exception and returns a boolean True if retries are permitted.
     :param session: Optional exiting requests.Session instance to reuse.
     :param infinite: Bool, set True to retry indefinitely. Default: False.
     :param log_req_resp: Set False to turn off verbose debug messages.
@@ -519,12 +263,19 @@ def readurl(
     if retries:
         manual_tries = max(int(retries) + 1, 1)
 
-    user_agent = "Cloud-Init/%s" % (version.version_string())
-    if headers is not None:
-        headers = headers.copy()
-    else:
-        headers = {}
+    def_headers = {
+        "User-Agent": "Cloud-Init/%s" % (version.version_string()),
+    }
+    if headers:
+        def_headers.update(headers)
+    headers = def_headers
 
+    if not headers_cb:
+
+        def _cb(url):
+            return headers
+
+        headers_cb = _cb
     if data:
         req_args["data"] = data
     if sec_between is None:
@@ -536,15 +287,9 @@ def readurl(
     # Handle retrying ourselves since the built-in support
     # doesn't handle sleeping between tries...
     for i in count():
-        if headers_cb:
-            headers = headers_cb(url)
-
-        if "User-Agent" not in headers:
-            headers["User-Agent"] = user_agent
-
-        req_args["headers"] = headers
+        req_args["headers"] = headers_cb(url)
         filtered_req_args = {}
-        for k, v in req_args.items():
+        for (k, v) in req_args.items():
             if k == "data":
                 continue
             if k == "headers" and headers_redact:
@@ -555,7 +300,6 @@ def readurl(
                         filtered_req_args[k][key] = REDACTED
             else:
                 filtered_req_args[k] = v
-        raised_exception: Exception
         try:
             if log_req_resp:
                 LOG.debug(
@@ -566,57 +310,59 @@ def readurl(
                     filtered_req_args,
                 )
 
-            response = session.request(**req_args)
+            r = session.request(**req_args)
 
             if check_status:
-                response.raise_for_status()
+                r.raise_for_status()
             LOG.debug(
                 "Read from %s (%s, %sb) after %s attempts",
                 url,
-                response.status_code,
-                len(response.content),
+                r.status_code,
+                len(r.content),
                 (i + 1),
             )
             # Doesn't seem like we can make it use a different
             # subclass for responses, so add our own backward-compat
             # attrs
-            return UrlResponse(response)
+            return UrlResponse(r)
         except exceptions.SSLError as e:
             # ssl exceptions are not going to get fixed by waiting a
             # few seconds
             raise UrlError(e, url=url) from e
-        except exceptions.HTTPError as e:
-            url_error = UrlError(
-                e,
-                code=e.response.status_code,
-                headers=e.response.headers,
-                url=url,
-            )
-            raised_exception = e
         except exceptions.RequestException as e:
-            url_error = UrlError(e, url=url)
-            raised_exception = e
-            response = None
+            if (
+                isinstance(e, (exceptions.HTTPError))
+                and hasattr(e, "response")
+                and hasattr(  # This appeared in v 0.10.8
+                    e.response, "status_code"
+                )
+            ):
+                url_error = UrlError(
+                    e,
+                    code=e.response.status_code,
+                    headers=e.response.headers,
+                    url=url,
+                )
+            else:
+                url_error = UrlError(e, url=url)
+
+            if exception_cb and not exception_cb(req_args.copy(), url_error):
+                # if an exception callback was given, it should return True
+                # to continue retrying and False to break and re-raise the
+                # exception
+                raise url_error from e
 
-        response_sleep_time = _handle_error(
-            url_error,
-            exception_cb=exception_cb,
-        )
-        # If our response tells us to wait, then wait even if we're
-        # past the max tries
-        if not response_sleep_time:
             will_retry = infinite or (i + 1 < manual_tries)
             if not will_retry:
-                raise url_error from raised_exception
-        sleep_time = response_sleep_time or sec_between
+                raise url_error from e
 
-        if sec_between > 0:
-            if log_req_resp:
-                LOG.debug(
-                    "Please wait %s seconds while we wait to try again",
-                    sec_between,
-                )
-            time.sleep(sleep_time)
+            if sec_between > 0:
+                if log_req_resp:
+                    LOG.debug(
+                        "Please wait %s seconds while we wait to try again",
+                        sec_between,
+                    )
+                time.sleep(sec_between)
 
     raise RuntimeError("This path should be unreachable...")
 
@@ -644,7 +390,7 @@ def dual_stack(
     addresses: List[str],
     stagger_delay: float = 0.150,
     timeout: int = 10,
-) -> Tuple[Optional[str], Optional[UrlResponse]]:
+) -> Tuple:
     """execute multiple callbacks in parallel
 
     Run blocking func against two different addresses staggered with a
@@ -654,7 +400,7 @@ def dual_stack(
     """
     return_result = None
     returned_address = None
-    last_exception: Optional[BaseException] = None
+    last_exception = None
     exceptions = []
     is_done = threading.Event()
 
@@ -714,7 +460,7 @@ def dual_stack(
             "Timed out waiting for addresses: %s, "
             "exception(s) raised while waiting: %s",
             " ".join(addresses),
-            " ".join(map(str, exceptions)),
+            " ".join(exceptions),  # type: ignore
         )
     finally:
         executor.shutdown(wait=False)
@@ -722,60 +468,61 @@ def dual_stack(
     return (returned_address, return_result)
 
 
-class HandledResponse(NamedTuple):
-    # Set when we have a response to return
-    url: Optional[str]
-    response: Optional[UrlResponse]
-
-    # Possibly set if we need to try again
-    wait_time: Optional[float]
-
-
 def wait_for_url(
     urls,
-    *,
     max_wait: float = float("inf"),
     timeout: Optional[float] = None,
     status_cb: Callable = LOG.debug,  # some sources use different log levels
     headers_cb: Optional[Callable] = None,
     headers_redact=None,
     sleep_time: Optional[float] = None,
-    exception_cb: ExceptionCallback = None,
+    exception_cb: Optional[Callable] = None,
     sleep_time_cb: Optional[Callable[[Any, float], float]] = None,
     request_method: str = "",
     connect_synchronously: bool = True,
     async_delay: float = 0.150,
 ):
-    """Wait for a response from one of the urls provided.
+    """
+    urls:      a list of urls to try
+    max_wait:  roughly the maximum time to wait before giving up
+               The max time is *actually* len(urls)*timeout as each url will
+               be tried once and given the timeout provided.
+               a number <= 0 will always result in only one try
+    timeout:   the timeout provided to urlopen
+    status_cb: call method with string message when a url is not available
+    headers_cb: call method with single argument of url to get headers
+                for request.
+    headers_redact: a list of header names to redact from the log
+    sleep_time: Amount of time to sleep between retries. If this and
+                sleep_time_cb are None, the default sleep time
+                defaults to 1 second and increases by 1 seconds every 5
+                tries. Cannot be specified along with `sleep_time_cb`.
+    exception_cb: call method with 2 arguments 'msg' (per status_cb) and
+                  'exception', the exception that occurred.
+    sleep_time_cb: call method with 2 arguments (response, loop_n) that
+                   generates the next sleep time. Cannot be specified
+                   along with 'sleep_time`.
+    request_method: indicate the type of HTTP request, GET, PUT, or POST
+    connect_synchronously: if false, enables executing requests in parallel
+    async_delay: delay before parallel metadata requests, see RFC 6555
+    returns: tuple of (url, response contents), on failure, (False, None)
 
-    :param urls: List of urls to try
-    :param max_wait: Roughly the maximum time to wait before giving up
-        The max time is *actually* len(urls)*timeout as each url will
-        be tried once and given the timeout provided.
-        a number <= 0 will always result in only one try
-    :param timeout: Timeout provided to urlopen
-    :param status_cb: Callable with string message when a url is not available
-    :param headers_cb: Callable with single argument of url to get headers
-        for request.
-    :param headers_redact: List of header names to redact from the log
-    :param sleep_time: Amount of time to sleep between retries. If this and
-        sleep_time_cb are None, the default sleep time defaults to 1 second
-        and increases by 1 seconds every 5 tries. Cannot be specified along
-        with `sleep_time_cb`.
-    :param exception_cb: Callable to handle exception and returns True if
-        retries are permitted.
-    :param sleep_time_cb: Callable with 2 arguments (response, loop_n) that
-        generates the next sleep time. Cannot be specified
-        along with 'sleep_time`.
-    :param request_method: Indicates the type of HTTP request:
-        GET, PUT, or POST
-    :param connect_synchronously: If false, enables executing requests
-        in parallel
-    :param async_delay: Delay before parallel metadata requests, see RFC 6555
+    the idea of this routine is to wait for the EC2 metadata service to
+    come up.  On both Eucalyptus and EC2 we have seen the case where
+    the instance hit the MD before the MD service was up.  EC2 seems
+    to have permanently fixed this, though.
 
-    :return: tuple of (url, response contents), on failure, (False, None)
+    In openstack, the metadata service might be painfully slow, and
+    unable to avoid hitting a timeout of even up to 10 seconds or more
+    (LP: #894279) for a simple GET.
 
-    :raises: UrlError on unrecoverable error
+    Offset those needs with the need to not hang forever (and block boot)
+    on a system where cloud-init is configured to look for EC2 Metadata
+    service but is not going to find one.  It is possible that the instance
+    data host (169.254.169.254) may be firewalled off Entirely for a system,
+    meaning that the connection will block forever unless a timeout is set.
+
+    The default value for max_wait will retry indefinitely.
     """
 
     def default_sleep_time(_, loop_number: int) -> float:
@@ -786,31 +533,11 @@ def wait_for_url(
         if max_wait in (float("inf"), None):
             return False
         return (max_wait <= 0) or (
-            time.monotonic() - start_time + sleep_time > max_wait
+            time.time() - start_time + sleep_time > max_wait
         )
 
-    def handle_url_response(
-        response: Optional[UrlResponse], url: Optional[str]
-    ) -> Tuple[Optional[UrlError], str]:
+    def handle_url_response(response, url):
         """Map requests response code/contents to internal "UrlError" type"""
-        reason = ""
-        url_exc = None
-        if not (response and url):
-            reason = "Request timed out"
-            url_exc = UrlError(ValueError(reason))
-            return url_exc, reason
-        try:
-            # Do this first because it can provide more context for the
-            # exception than what comes later
-            response._response.raise_for_status()
-        except requests.exceptions.HTTPError as e:
-            url_exc = UrlError(
-                e,
-                code=e.response.status_code,
-                headers=e.response.headers,
-                url=url,
-            )
-            return url_exc, str(e)
         if not response.contents:
             reason = "empty response [%s]" % (response.code)
             url_exc = UrlError(
@@ -820,7 +547,6 @@ def wait_for_url(
                 url=url,
             )
         elif not response.ok():
-            # 3xx "errors" wouldn't be covered by the raise_for_status above
             reason = "bad status code [%s]" % (response.code)
             url_exc = UrlError(
                 ValueError(reason),
@@ -828,53 +554,44 @@ def wait_for_url(
                 headers=response.headers,
                 url=url,
             )
+        else:
+            reason = ""
+            url_exc = None
         return (url_exc, reason)
 
     def read_url_handle_exceptions(
-        url_reader_cb: Callable[
-            [Any], Tuple[Optional[str], Optional[UrlResponse]]
-        ],
-        urls: Union[str, List[str]],
-        start_time: int,
-        exc_cb: ExceptionCallback,
-        log_cb: Callable,
-    ) -> HandledResponse:
+        url_reader_cb, urls, start_time, exc_cb, log_cb
+    ):
         """Execute request, handle response, optionally log exception"""
         reason = ""
         url = None
-        url_exc: Optional[Exception]
         try:
             url, response = url_reader_cb(urls)
             url_exc, reason = handle_url_response(response, url)
             if not url_exc:
-                return HandledResponse(url, response, wait_time=None)
+                return (url, response)
         except UrlError as e:
             reason = "request error [%s]" % e
             url_exc = e
         except Exception as e:
             reason = "unexpected error [%s]" % e
             url_exc = e
-        time_taken = int(time.monotonic() - start_time)
+        time_taken = int(time.time() - start_time)
         max_wait_str = "%ss" % max_wait if max_wait else "unlimited"
         status_msg = "Calling '%s' failed [%s/%s]: %s" % (
-            url or getattr(url_exc, "url", "url"),
+            url or getattr(url_exc, "url", "url ? None"),
             time_taken,
             max_wait_str,
             reason,
         )
         log_cb(status_msg)
+        if exc_cb:
+            # This can be used to alter the headers that will be sent
+            # in the future, for example this is what the MAAS datasource
+            # does.
+            exc_cb(msg=status_msg, exception=url_exc)
 
-        return HandledResponse(
-            url=None,
-            response=None,
-            wait_time=(
-                _handle_error(url_exc, exception_cb=exc_cb)
-                if isinstance(url_exc, UrlError)
-                else None
-            ),
-        )
-
-    def read_url_cb(url: str, timeout: int) -> UrlResponse:
+    def read_url_cb(url, timeout):
         return readurl(
             url,
             headers={} if headers_cb is None else headers_cb(url),
@@ -884,24 +601,19 @@ def wait_for_url(
             request_method=request_method,
         )
 
-    def read_url_serial(
-        start_time, timeout, exc_cb, log_cb
-    ) -> HandledResponse:
+    def read_url_serial(start_time, timeout, exc_cb, log_cb):
         """iterate over list of urls, request each one and handle responses
         and thrown exceptions individually per url
         """
 
-        def url_reader_serial(url: str):
+        def url_reader_serial(url):
             return (url, read_url_cb(url, timeout))
 
-        wait_times = []
         for url in urls:
-            now = time.monotonic()
-            if loop_n != 0 and not must_try_again:
+            now = time.time()
+            if loop_n != 0:
                 if timeup(max_wait, start_time):
-                    return HandledResponse(
-                        url=None, response=None, wait_time=None
-                    )
+                    return
                 if (
                     max_wait is not None
                     and timeout
@@ -913,16 +625,10 @@ def wait_for_url(
             out = read_url_handle_exceptions(
                 url_reader_serial, url, start_time, exc_cb, log_cb
             )
-            if out.response:
+            if out:
                 return out
-            elif out.wait_time:
-                wait_times.append(out.wait_time)
-        wait_time = max(wait_times) if wait_times else None
-        return HandledResponse(url=None, response=None, wait_time=wait_time)
 
-    def read_url_parallel(
-        start_time, timeout, exc_cb, log_cb
-    ) -> HandledResponse:
+    def read_url_parallel(start_time, timeout, exc_cb, log_cb):
         """pass list of urls to dual_stack which sends requests in parallel
         handle response and exceptions of the first endpoint to respond
         """
@@ -932,11 +638,13 @@ def wait_for_url(
             stagger_delay=async_delay,
             timeout=timeout,
         )
-        return read_url_handle_exceptions(
+        out = read_url_handle_exceptions(
             url_reader_parallel, urls, start_time, exc_cb, log_cb
         )
+        if out:
+            return out
 
-    start_time = time.monotonic()
+    start_time = time.time()
     if sleep_time and sleep_time_cb:
         raise ValueError("sleep_time and sleep_time_cb are mutually exclusive")
 
@@ -954,17 +662,13 @@ def wait_for_url(
     loop_n: int = 0
     response = None
     while True:
-        resp = do_read_url(start_time, timeout, exception_cb, status_cb)
-        must_try_again = False
-        if resp.response:
-            return resp.url, resp.response.contents
-        elif resp.wait_time:
-            time.sleep(resp.wait_time)
-            loop_n = loop_n + 1
-            must_try_again = True
-            continue
-
         current_sleep_time = calculate_sleep_time(response, loop_n)
+
+        url = do_read_url(start_time, timeout, exception_cb, status_cb)
+        if url:
+            address, response = url
+            return (address, response.contents)
+
         if timeup(max_wait, start_time, current_sleep_time):
             break
 
@@ -976,13 +680,14 @@ def wait_for_url(
         time.sleep(current_sleep_time)
 
         # shorten timeout to not run way over max_time
-        current_time = time.monotonic()
+        current_time = time.time()
         if timeout and current_time + timeout > start_time + max_wait:
             timeout = max_wait - (current_time - start_time)
             if timeout <= 0:
                 # We've already exceeded our max_wait. Time to bail.
                 break
 
+    LOG.error("Timed out, no response from urls: %s", urls)
     return False, None
 
 
@@ -1016,9 +721,7 @@ class OauthUrlHelper:
 
     def read_skew_file(self):
         if self.skew_data_file and os.path.isfile(self.skew_data_file):
-            with performance.Timed(f"Reading {self.skew_data_file}"), open(
-                self.skew_data_file, mode="r"
-            ) as fp:
+            with open(self.skew_data_file, mode="r") as fp:
                 return json.load(fp)
         return None
 
@@ -1030,12 +733,10 @@ class OauthUrlHelper:
         if cur is None:
             cur = {}
         cur[host] = value
-        with performance.Timed(f"Writing {self.skew_data_file}"), open(
-            self.skew_data_file, mode="w"
-        ) as fp:
+        with open(self.skew_data_file, mode="w") as fp:
             fp.write(json.dumps(cur))
 
-    def exception_cb(self, exception):
+    def exception_cb(self, msg, exception):
         if not (
             isinstance(exception, UrlError)
             and (exception.code == 403 or exception.code == 401)
@@ -1096,13 +797,13 @@ class OauthUrlHelper:
     def readurl(self, *args, **kwargs):
         return self._wrapped(readurl, args, kwargs)
 
-    def _exception_cb(self, extra_exception_cb, exception):
+    def _exception_cb(self, extra_exception_cb, msg, exception):
         ret = None
         try:
             if extra_exception_cb:
-                ret = extra_exception_cb(exception)
+                ret = extra_exception_cb(msg, exception)
         finally:
-            self.exception_cb(exception)
+            self.exception_cb(msg, exception)
         return ret
 
     def _headers_cb(self, extra_headers_cb, url):
