@@ -3,10 +3,13 @@
 # for complete details.
 
 
+import abc
 import typing
 
 from cryptography.hazmat.primitives import hashes
-from cryptography.hazmat.primitives._asymmetric import AsymmetricPadding
+from cryptography.hazmat.primitives._asymmetric import (
+    AsymmetricPadding as AsymmetricPadding,
+)
 from cryptography.hazmat.primitives.asymmetric import rsa
 
 
@@ -14,20 +17,41 @@ class PKCS1v15(AsymmetricPadding):
     name = "EMSA-PKCS1-v1_5"
 
 
-class PSS(AsymmetricPadding):
-    MAX_LENGTH = object()
-    name = "EMSA-PSS"
+class _MaxLength:
+    "Sentinel value for `MAX_LENGTH`."
 
-    def __init__(self, mgf, salt_length):
+
+class _Auto:
+    "Sentinel value for `AUTO`."
+
+
+class _DigestLength:
+    "Sentinel value for `DIGEST_LENGTH`."
+
+
+class PSS(AsymmetricPadding):
+    MAX_LENGTH = _MaxLength()
+    AUTO = _Auto()
+    DIGEST_LENGTH = _DigestLength()
+    name = "EMSA-PSS"
+    _salt_length: typing.Union[int, _MaxLength, _Auto, _DigestLength]
+
+    def __init__(
+        self,
+        mgf: "MGF",
+        salt_length: typing.Union[int, _MaxLength, _Auto, _DigestLength],
+    ) -> None:
         self._mgf = mgf
 
-        if (
-            not isinstance(salt_length, int)
-            and salt_length is not self.MAX_LENGTH
+        if not isinstance(
+            salt_length, (int, _MaxLength, _Auto, _DigestLength)
         ):
-            raise TypeError("salt_length must be an integer.")
+            raise TypeError(
+                "salt_length must be an integer, MAX_LENGTH, "
+                "DIGEST_LENGTH, or AUTO"
+            )
 
-        if salt_length is not self.MAX_LENGTH and salt_length < 0:
+        if isinstance(salt_length, int) and salt_length < 0:
             raise ValueError("salt_length must be zero or greater.")
 
         self._salt_length = salt_length
@@ -38,7 +62,7 @@ class OAEP(AsymmetricPadding):
 
     def __init__(
         self,
-        mgf: "MGF1",
+        mgf: "MGF",
         algorithm: hashes.HashAlgorithm,
         label: typing.Optional[bytes],
     ):
@@ -50,8 +74,12 @@ class OAEP(AsymmetricPadding):
         self._label = label
 
 
-class MGF1(object):
-    MAX_LENGTH = object()
+class MGF(metaclass=abc.ABCMeta):
+    _algorithm: hashes.HashAlgorithm
+
+
+class MGF1(MGF):
+    MAX_LENGTH = _MaxLength()
 
     def __init__(self, algorithm: hashes.HashAlgorithm):
         if not isinstance(algorithm, hashes.HashAlgorithm):
