@@ -8,40 +8,32 @@ import ipaddress
 import typing
 from email.utils import parseaddr
 
-from cryptography import utils
 from cryptography.x509.name import Name
 from cryptography.x509.oid import ObjectIdentifier
 
 
-_GENERAL_NAMES = {
-    0: "otherName",
-    1: "rfc822Name",
-    2: "dNSName",
-    3: "x400Address",
-    4: "directoryName",
-    5: "ediPartyName",
-    6: "uniformResourceIdentifier",
-    7: "iPAddress",
-    8: "registeredID",
-}
+_IPADDRESS_TYPES = typing.Union[
+    ipaddress.IPv4Address,
+    ipaddress.IPv6Address,
+    ipaddress.IPv4Network,
+    ipaddress.IPv6Network,
+]
 
 
 class UnsupportedGeneralNameType(Exception):
-    def __init__(self, msg, type):
-        super(UnsupportedGeneralNameType, self).__init__(msg)
-        self.type = type
+    pass
 
 
 class GeneralName(metaclass=abc.ABCMeta):
     @abc.abstractproperty
-    def value(self):
+    def value(self) -> typing.Any:
         """
         Return the value of the object
         """
 
 
 class RFC822Name(GeneralName):
-    def __init__(self, value: str):
+    def __init__(self, value: str) -> None:
         if isinstance(value, str):
             try:
                 value.encode("ascii")
@@ -62,10 +54,12 @@ class RFC822Name(GeneralName):
 
         self._value = value
 
-    value = utils.read_only_property("_value")
+    @property
+    def value(self) -> str:
+        return self._value
 
     @classmethod
-    def _init_without_validation(cls, value):
+    def _init_without_validation(cls, value: str) -> "RFC822Name":
         instance = cls.__new__(cls)
         instance._value = value
         return instance
@@ -79,15 +73,12 @@ class RFC822Name(GeneralName):
 
         return self.value == other.value
 
-    def __ne__(self, other: object) -> bool:
-        return not self == other
-
     def __hash__(self) -> int:
         return hash(self.value)
 
 
 class DNSName(GeneralName):
-    def __init__(self, value: str):
+    def __init__(self, value: str) -> None:
         if isinstance(value, str):
             try:
                 value.encode("ascii")
@@ -102,15 +93,17 @@ class DNSName(GeneralName):
 
         self._value = value
 
-    value = utils.read_only_property("_value")
+    @property
+    def value(self) -> str:
+        return self._value
 
     @classmethod
-    def _init_without_validation(cls, value):
+    def _init_without_validation(cls, value: str) -> "DNSName":
         instance = cls.__new__(cls)
         instance._value = value
         return instance
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "<DNSName(value={0!r})>".format(self.value)
 
     def __eq__(self, other: object) -> bool:
@@ -119,15 +112,12 @@ class DNSName(GeneralName):
 
         return self.value == other.value
 
-    def __ne__(self, other: object) -> bool:
-        return not self == other
-
     def __hash__(self) -> int:
         return hash(self.value)
 
 
 class UniformResourceIdentifier(GeneralName):
-    def __init__(self, value: str):
+    def __init__(self, value: str) -> None:
         if isinstance(value, str):
             try:
                 value.encode("ascii")
@@ -142,10 +132,14 @@ class UniformResourceIdentifier(GeneralName):
 
         self._value = value
 
-    value = utils.read_only_property("_value")
+    @property
+    def value(self) -> str:
+        return self._value
 
     @classmethod
-    def _init_without_validation(cls, value):
+    def _init_without_validation(
+        cls, value: str
+    ) -> "UniformResourceIdentifier":
         instance = cls.__new__(cls)
         instance._value = value
         return instance
@@ -159,21 +153,20 @@ class UniformResourceIdentifier(GeneralName):
 
         return self.value == other.value
 
-    def __ne__(self, other: object) -> bool:
-        return not self == other
-
     def __hash__(self) -> int:
         return hash(self.value)
 
 
 class DirectoryName(GeneralName):
-    def __init__(self, value: Name):
+    def __init__(self, value: Name) -> None:
         if not isinstance(value, Name):
             raise TypeError("value must be a Name")
 
         self._value = value
 
-    value = utils.read_only_property("_value")
+    @property
+    def value(self) -> Name:
+        return self._value
 
     def __repr__(self) -> str:
         return "<DirectoryName(value={})>".format(self.value)
@@ -184,21 +177,20 @@ class DirectoryName(GeneralName):
 
         return self.value == other.value
 
-    def __ne__(self, other: object) -> bool:
-        return not self == other
-
     def __hash__(self) -> int:
         return hash(self.value)
 
 
 class RegisteredID(GeneralName):
-    def __init__(self, value: ObjectIdentifier):
+    def __init__(self, value: ObjectIdentifier) -> None:
         if not isinstance(value, ObjectIdentifier):
             raise TypeError("value must be an ObjectIdentifier")
 
         self._value = value
 
-    value = utils.read_only_property("_value")
+    @property
+    def value(self) -> ObjectIdentifier:
+        return self._value
 
     def __repr__(self) -> str:
         return "<RegisteredID(value={})>".format(self.value)
@@ -209,23 +201,12 @@ class RegisteredID(GeneralName):
 
         return self.value == other.value
 
-    def __ne__(self, other: object) -> bool:
-        return not self == other
-
     def __hash__(self) -> int:
         return hash(self.value)
 
 
 class IPAddress(GeneralName):
-    def __init__(
-        self,
-        value: typing.Union[
-            ipaddress.IPv4Address,
-            ipaddress.IPv6Address,
-            ipaddress.IPv4Network,
-            ipaddress.IPv6Network,
-        ],
-    ):
+    def __init__(self, value: _IPADDRESS_TYPES) -> None:
         if not isinstance(
             value,
             (
@@ -243,7 +224,19 @@ class IPAddress(GeneralName):
 
         self._value = value
 
-    value = utils.read_only_property("_value")
+    @property
+    def value(self) -> _IPADDRESS_TYPES:
+        return self._value
+
+    def _packed(self) -> bytes:
+        if isinstance(
+            self.value, (ipaddress.IPv4Address, ipaddress.IPv6Address)
+        ):
+            return self.value.packed
+        else:
+            return (
+                self.value.network_address.packed + self.value.netmask.packed
+            )
 
     def __repr__(self) -> str:
         return "<IPAddress(value={})>".format(self.value)
@@ -254,15 +247,12 @@ class IPAddress(GeneralName):
 
         return self.value == other.value
 
-    def __ne__(self, other: object) -> bool:
-        return not self == other
-
     def __hash__(self) -> int:
         return hash(self.value)
 
 
 class OtherName(GeneralName):
-    def __init__(self, type_id: ObjectIdentifier, value: bytes):
+    def __init__(self, type_id: ObjectIdentifier, value: bytes) -> None:
         if not isinstance(type_id, ObjectIdentifier):
             raise TypeError("type_id must be an ObjectIdentifier")
         if not isinstance(value, bytes):
@@ -271,8 +261,13 @@ class OtherName(GeneralName):
         self._type_id = type_id
         self._value = value
 
-    type_id = utils.read_only_property("_type_id")
-    value = utils.read_only_property("_value")
+    @property
+    def type_id(self) -> ObjectIdentifier:
+        return self._type_id
+
+    @property
+    def value(self) -> bytes:
+        return self._value
 
     def __repr__(self) -> str:
         return "<OtherName(type_id={}, value={!r})>".format(
@@ -285,8 +280,5 @@ class OtherName(GeneralName):
 
         return self.type_id == other.type_id and self.value == other.value
 
-    def __ne__(self, other: object) -> bool:
-        return not self == other
-
     def __hash__(self) -> int:
         return hash((self.type_id, self.value))
