@@ -1,24 +1,22 @@
 """
 Client version related functions
 """
+
 import os.path
-import re
 from math import inf
 from typing import Optional
 
 from uaclient.apt import (
-    compare_versions,
-    get_apt_cache_policy_for_package,
     get_apt_cache_time,
+    get_pkg_candidate_version,
+    version_compare,
 )
 from uaclient.defaults import CANDIDATE_CACHE_PATH, UAC_RUN_PATH
 from uaclient.exceptions import ProcessExecutionError
 from uaclient.system import subp
 
-__VERSION__ = "27.14.4"
-PACKAGED_VERSION = "27.14.4"
-
-CANDIDATE_REGEX = r"Candidate: (?P<candidate>.*?)\n"
+__VERSION__ = "32.3.1"
+PACKAGED_VERSION = "32.3.1~23.10"
 
 
 def get_version() -> str:
@@ -56,10 +54,8 @@ def get_last_known_candidate() -> Optional[str]:
     ):
         candidate_version = None
         try:
-            policy = get_apt_cache_policy_for_package("ubuntu-advantage-tools")
-            match = re.search(CANDIDATE_REGEX, policy)
-            if match:
-                candidate_version = match.group("candidate")
+            candidate_version = get_pkg_candidate_version("ubuntu-pro-client")
+            if candidate_version:
                 os.makedirs(UAC_RUN_PATH, exist_ok=True)
                 with open(CANDIDATE_CACHE_PATH, "w") as f:
                     f.write(candidate_version)
@@ -79,6 +75,6 @@ def get_last_known_candidate() -> Optional[str]:
 
 def check_for_new_version() -> Optional[str]:
     candidate = get_last_known_candidate()
-    if candidate and compare_versions(candidate, get_version(), "gt"):
+    if candidate and version_compare(candidate, get_version()) > 0:
         return candidate
     return None
