@@ -1,15 +1,18 @@
 """Send reports about ubuntu-desktop-installer to the correct Launchpad
 project."""
+
+# pylint: disable=invalid-name
+# pylint: enable=invalid-name
+
 import os
 
 from apport import hookutils
 
 
 def add_info(report, unused_ui):
-
-    udilog = os.path.realpath(
-        "/var/log/installer/ubuntu_desktop_installer.log"
-    )
+    """Send reports about ubuntu-desktop-installer to the correct Launchpad
+    project."""
+    udilog = os.path.realpath("/var/log/installer/ubuntu_desktop_installer.log")
     hookutils.attach_file_if_exists(report, udilog, "UdiLog")
 
     report["SourcePackage"] = "ubuntu-desktop-installer"
@@ -25,19 +28,13 @@ def add_info(report, unused_ui):
 }
 """
 
-    subiquitylog = os.path.realpath(
-        "/var/log/installer/subiquity-server-debug.log"
-    )
+    subiquitylog = os.path.realpath("/var/log/installer/subiquity-server-debug.log")
     hookutils.attach_file_if_exists(report, subiquitylog, "SubiquityLog")
 
     hookutils.attach_file_if_exists(
-        report,
-        "/var/log/installer/subiquity-curtin-install.conf",
-        "CurtinConfig",
-    )
-    hookutils.attach_file_if_exists(
-        report, "/var/log/curtin/install.log", "CurtinLog"
+        report, "/var/log/installer/subiquity-curtin-install.conf", "CurtinConfig"
     )
+    hookutils.attach_file_if_exists(report, "/var/log/curtin/install.log", "CurtinLog")
     hookutils.attach_file_if_exists(
         report, "/var/log/curtin/curtin-error-logs.tar", "CurtinError"
     )
